package tools.descartes.prisma.agent.generic;

import java.io.File;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.eclipse.emf.cdo.CDOObject;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature.Setting;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.equinox.app.IApplication;
import org.eclipse.equinox.app.IApplicationContext;

import edu.kit.ipd.descartes.identifier.Identifier;
import tools.descartes.prisma.agent.generic.genericagent.GenericAgentConfiguration;
import tools.descartes.prisma.agent.generic.genericagent.RequiredObject;
import tools.descartes.prisma.core.ModelRepository;
import tools.descartes.prisma.core.adapter.ModelSkeletonAdapter;
import tools.descartes.prisma.core.adapter.ObjectMatcher;
import tools.descartes.prisma.core.agent.AgentController;
import tools.descartes.prisma.core.agent.DelegationProvider;
import tools.descartes.prisma.core.impl.CDOModelRepository;
import tools.descartes.prisma.model.scope.DelegatedObject;
import tools.descartes.prisma.model.skeleton.ModelSkeleton;

public class GenericAgent implements IApplication {

	private class GenericAgentController extends AgentController {

		private ModelSkeleton skeleton;

		private final Set<EObject> delegatedObjects = new HashSet<>();

		public GenericAgentController(ModelRepository repository, GenericAgentConfiguration configuration) {
			super(repository, configuration.getScope().getPath(), configuration.getId());
		}

		@Override
		protected void notifyAddDelegatedElement(EObject newElement, Map<String, String> additionalProperties) {
			delegatedObjects.add(newElement);
		}

		@Override
		protected void notifyRemoveDelegatedElement(EObject oldElement, Map<String, String> additionalProperties) {
			delegatedObjects.remove(oldElement);
		}

		public Set<EObject> getDelegatedObjects() {
			return delegatedObjects;
		}

	}

	private class GenericDelegationProvider implements DelegationProvider {

		private final Map<String, Map<String, String>> elements = new HashMap<>();

		public GenericDelegationProvider(GenericAgentConfiguration configuration) {
			for (DelegatedObject curElement : configuration.getDelegatedObjects()) {
				EObject object = curElement.getObject();
				if (object instanceof Identifier) {
					elements.put(((Identifier) object).getId(), curElement.getAdditionalProperties().map());
				}
			}
		}

		@Override
		public boolean needsDelegation(EObject object) {
			if (object instanceof Identifier) {
				return elements.containsKey(((Identifier) object).getId());
			}
			return false;
		}

		@Override
		public Map<String, String> getAdditionalProperties(EObject object) {
			if (object instanceof Identifier) {
				Map<String, String> properties = elements.get(((Identifier) object).getId());
				if (properties != null) {
					return properties;
				}
			}
			return Collections.emptyMap();
		}

	}

	@Override
	public Object start(IApplicationContext context) throws Exception {
		String[] args = (String[]) context.getArguments().get(IApplicationContext.APPLICATION_ARGS);

		Options options = new Options();
		options.addOption("r", "repository", true, "Host of the model repository");
		options.addOption("p", "port", true, "Port of the model repository");
		options.addOption("c", "config", true, "Agent configuration");

		CommandLineParser cliParser = new GnuParser();
		CommandLine cli = cliParser.parse(options, args);

		String repositoryURL = null;
		String configurationPath = null;
		boolean help = false;
		if (cli.hasOption("r")) {
			repositoryURL = cli.getOptionValue("r");
		} else {
			help = true;
		}
		if (cli.hasOption("c")) {
			configurationPath = cli.getOptionValue("c");
		} else {
			help = true;
		}

		if (help) {
			HelpFormatter helpFormat = new HelpFormatter();
			helpFormat.printHelp("genericagent", options);
			return null;
		}
		GenericAgentConfiguration configuration = loadConfiguration(configurationPath);

		ModelRepository repository = new CDOModelRepository(repositoryURL);
		GenericAgentController agent = new GenericAgentController(repository, configuration);

		agent.setDelegationProvider(new GenericDelegationProvider(configuration));

		agent.start().get();

		if (resolveRequiredObjects(agent, configuration)) {
			agent.apply(configuration.getSkeleton());
		}
		
		agent.stop().get();

		return null;
	}

	@Override
	public void stop() {
		// TODO Auto-generated method stub
	}

	private GenericAgentConfiguration loadConfiguration(String configurationPath) {
		ResourceSet set = new ResourceSetImpl();
		Resource res = set.getResource(URI.createFileURI(new File(configurationPath).getAbsolutePath()), true);
		GenericAgentConfiguration configuration = (GenericAgentConfiguration) res.getContents().get(0);

		// IMPORTANT: resolve all cross-references
		EcoreUtil.resolveAll(set);
		Map<EObject, Collection<Setting>> unresolved = EcoreUtil.UnresolvedProxyCrossReferencer.find(set);
		for (EObject problemObject : unresolved.keySet()) {
			Collection<Setting> unresolvedSettings = unresolved.get(problemObject);
			for (Setting curSetting : unresolvedSettings) {
				System.out.println("Unresolved proxy: " + curSetting.getEStructuralFeature());
			}
		}

		return configuration;
	}


	private boolean resolveRequiredObjects(GenericAgentController agent, GenericAgentConfiguration configuration) {
		Set<RequiredObject> notResolved = new HashSet<>(configuration.getRequiredObjects());
		for (EObject curDelegate : agent.getDelegatedObjects()) {
			for (RequiredObject reqObject : notResolved) {
				if (ObjectMatcher.matches(curDelegate, reqObject.getObject())) {
					// Connect the two objects with each other
					ModelSkeletonAdapter adapter = agent.getModelRepository().adapt(reqObject.getObject(),
							ModelSkeletonAdapter.class);
					adapter.setRepositoryObject((CDOObject) curDelegate);

					// Requirement is fulfilled
					notResolved.remove(reqObject);
					break;
				}
			}
		}

		if (!notResolved.isEmpty()) {
			for (RequiredObject unresolvedObject : notResolved) {
				System.err.println("Could not resolve required object: " + unresolvedObject.getObject());
			}
			return false;
		}
		return true;
	}

}
