package tools.descartes.prisma.parameterization;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import tools.descartes.librede.configuration.ExternalCall;
import tools.descartes.librede.configuration.ModelEntity;
import tools.descartes.librede.configuration.Service;
import tools.descartes.librede.metrics.Aggregation;
import tools.descartes.librede.metrics.Metric;
import tools.descartes.librede.metrics.StandardMetrics;
import tools.descartes.librede.repository.IMonitoringRepository;
import tools.descartes.librede.repository.handlers.BaseDerivationHandler;
import tools.descartes.librede.repository.handlers.DeriveIdentityHandler;
import tools.descartes.librede.repository.rules.Rule;
import tools.descartes.librede.repository.rules.RulePrecondition;
import tools.descartes.librede.repository.rules.RuleScope;
import tools.descartes.librede.units.Quantity;
import tools.descartes.librede.units.RequestCount;
import tools.descartes.librede.units.Time;
import tools.descartes.librede.units.Unit;

public class WorkOrderServiceProxyWorkaround {
	
	private static class ProxyServicePrecondition implements RulePrecondition {

		@Override
		public boolean check(ModelEntity entity) {
			if (entity instanceof Service) {
				return entity.getName().startsWith("Proxy:") && ((Service)entity).getOutgoingCalls().size() == 1;
			}
			return false;
		}
		
	}
	
	private static class ProxyExternalCallPrecondition implements RulePrecondition {

		@Override
		public boolean check(ModelEntity entity) {
			if (entity instanceof ExternalCall) {
				return ((ExternalCall)entity).getService().getName().startsWith("Proxy:");
			}
			if (entity instanceof Service) {
				return ((Service)entity).getName().startsWith("Proxy:");
			}
			return false;
		}
		
	}
	
	private static class ProxyScope implements RuleScope {

		@Override
		public Set<? extends ModelEntity> getScopeSet(ModelEntity base) {
			if (base instanceof ExternalCall) {
				ExternalCall c = (ExternalCall)base;
				Set<ModelEntity> entities = new HashSet<>();
				entities.add(c.getCalledService());
				entities.addAll(c.getCalledService().getIncomingCalls());
				entities.remove(c); // Remove myself from scope set
				return entities;
			}
			return Collections.emptySet();
		}

		@Override
		public Set<? extends ModelEntity> getNotificationSet(ModelEntity base) {
			if (base instanceof Service) {
				return new HashSet<>(((Service) base).getIncomingCalls());
			}
			if (base instanceof ExternalCall) {
				Set<ModelEntity> entities = new HashSet<>(((ExternalCall)base).getCalledService().getIncomingCalls());
				entities.remove(base);
				return entities;
			}
			return Collections.emptySet();
		}
		
	}
	
	private static class ProxyDeparturesDerivation extends BaseDerivationHandler<RequestCount> {
		@Override
		public double aggregate(IMonitoringRepository repository, Metric<RequestCount> metric, Unit<RequestCount> unit,
				ModelEntity entity, Aggregation aggregation, Quantity<Time> start, Quantity<Time> end) {
			if (entity instanceof ExternalCall && aggregation == Aggregation.SUM) {
				ExternalCall proxyCall = (ExternalCall)entity;
				double departures = repository.aggregate(metric, unit, proxyCall.getCalledService(), aggregation, start, end);
				for (ExternalCall c : proxyCall.getCalledService().getIncomingCalls()) {
					if (c != proxyCall) {
						departures -= repository.aggregate(metric, unit, c, aggregation, start, end);
					}
				}
				return departures;
			}
			throw new IllegalArgumentException();
		}
	}
	
	private static class CopyDeparturesDerivation extends BaseDerivationHandler<RequestCount> {
		@Override
		public double aggregate(IMonitoringRepository repository, Metric<RequestCount> metric, Unit<RequestCount> unit,
				ModelEntity entity, Aggregation aggregation, Quantity<Time> start, Quantity<Time> end) {
			if (entity instanceof Service && aggregation == Aggregation.SUM) {
				Service curService = (Service)entity;
				return repository.aggregate(metric, unit, curService.getOutgoingCalls().get(0), aggregation, start, end);
			}
			throw new IllegalArgumentException();
		}
	}
	
	private static class CopyResponseTimesDerivation extends BaseDerivationHandler<Time> {
		@Override
		public double aggregate(IMonitoringRepository repository, Metric<Time> metric, Unit<Time> unit,
				ModelEntity entity, Aggregation aggregation, Quantity<Time> start, Quantity<Time> end) {
			if (entity instanceof Service && aggregation == Aggregation.SUM) {
				Service curService = (Service)entity;
				Service proxied = curService.getOutgoingCalls().get(0).getCalledService();
				return repository.aggregate(metric, unit, proxied, aggregation, start, end);
			}
			throw new IllegalArgumentException();
		}
	}
	
	public static List<Rule<?>> getRules() {
		return Arrays.asList(		
				Rule.rule(StandardMetrics.DEPARTURES, Aggregation.SUM)
				.requiring(StandardMetrics.DEPARTURES, Aggregation.SUM)
				.check(new ProxyExternalCallPrecondition())
				.scope(new ProxyScope())
				.build(new ProxyDeparturesDerivation()),
				Rule.rule(StandardMetrics.DEPARTURES, Aggregation.SUM)
				.requiring(StandardMetrics.DEPARTURES, Aggregation.SUM)
				.check(new ProxyServicePrecondition())
				.scope(new RuleScope() {
					
					@Override
					public Set<? extends ModelEntity> getScopeSet(ModelEntity base) {
						if (base instanceof Service) {
							return new HashSet<>(((Service)base).getOutgoingCalls());
						}
						return Collections.emptySet();
					}
					
					@Override
					public Set<? extends ModelEntity> getNotificationSet(ModelEntity base) {
						if (base instanceof ExternalCall) {
							return Collections.singleton(((ExternalCall)base).getService());
						}
						return Collections.emptySet();
					}
				})
				.build(new CopyDeparturesDerivation())
			);
	}
	
	public static boolean isProxyService(Service curService) {
		return curService.getName().startsWith("Proxy:");
	}

}
